/**
  *  \file server/objcmd.cc
  *  \brief Server: Object Commands
  */

#include "server/objcmd.h"
#include "int/interror.h"
#include "server/objout.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "game/global.h"
#include "int/args.h"
#include "int/if/ifutil.h"
#include "game/ship.h"
#include "game/mission.h"
#include "int/exec.h"
#include "int/opcode.h"
#include "int/bytecode.h"
#include "int/if/baseif.h"

/** Execution context for commands.

    Some commands can be implemented as a setXxxProperty call. For others, there
    are two ways:
    - generate a temporary IntExecutionContext containing the args and context,
      and a temporary BCO containing "pushmem", "callind" instructions to look
      up and call the method by name.
    - call the respective implementation directly.

    On my computer, the former takes ten times as many cycles than the former
    for a "setwaypoint" call (60000 vs. 6000). Therefore, there's a strong
    preference for the second variant. To avoid having to export all methods
    from shipif/planetif, I added callShipMethod/callPlanetMethod in the same
    style as setShipProperty/setPlanetProperty.

    The need an execution context, which they would use to place return values
    in ("CARGO.REMAINDER"). Errors are still reported through C++ exceptions,
    which normally the interpreter loop would catch; here, the POST implementation
    catches them. */
static IntExecutionContext serverXC("server");

ServerObjectCommandHandler::~ServerObjectCommandHandler()
{ }


/*********************** ServerPlanetCommandHandler **********************/

ServerPlanetCommandHandler::ServerPlanetCommandHandler(int pid)
    : pid(pid)
{ }

void
ServerPlanetCommandHandler::processCommand(string_t cmd, IntArgBlock& args, ptr_vector<ServerObjectWriter>& result)
{
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    GPlanet& pl = univ.getPlanet(pid);
    // Missing: mark, unmark
    if (cmd == "setcomment") {
        callPlanetMethod(pl, univ, ipmSetComment, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "setfcode") {
        args.checkArgumentCount(1);
        setPlanetProperty(pl, ippFCode, args.getNext());
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "fixship") {
        callPlanetMethod(pl, univ, ipmFixShip, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "recycleship") {
        callPlanetMethod(pl, univ, ipmRecycleShip, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildbase") {
        callPlanetMethod(pl, univ, ipmBuildBase, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "autobuild") {
        callPlanetMethod(pl, univ, ipmAutoBuild, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "builddefense") {
        callPlanetMethod(pl, univ, ipmBuildDefense, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildfactories") {
        callPlanetMethod(pl, univ, ipmBuildFactories, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildmines") {
        callPlanetMethod(pl, univ, ipmBuildMines, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildbasedefense") {
        callPlanetMethod(pl, univ, ipmBuildBaseDefense, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "setcolonisttax") {
        args.checkArgumentCount(1);
        setPlanetProperty(pl, ippColonistTax, args.getNext());
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "setnativetax") {
        args.checkArgumentCount(1);
        setPlanetProperty(pl, ippNativeTax, args.getNext());
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "setmission") {
        args.checkArgumentCount(1);
        setBaseProperty(pl, ibpMission, args.getNext());
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "settech") {
        callPlanetMethod(pl, univ, ipmSetTech, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildfighters") {
        callPlanetMethod(pl, univ, ipmBuildFighters, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildengines") {
        callPlanetMethod(pl, univ, ipmBuildEngines, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildtorps") {
        callPlanetMethod(pl, univ, ipmBuildTorps, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildhulls") {
        callPlanetMethod(pl, univ, ipmBuildHulls, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildlaunchers") {
        callPlanetMethod(pl, univ, ipmBuildLaunchers, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildbeams") {
        callPlanetMethod(pl, univ, ipmBuildBeams, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "sellsupplies") {
        callPlanetMethod(pl, univ, ipmSellSupplies, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "buildship") {
        callPlanetMethod(pl, univ, ipmBuildShip, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
    } else if (cmd == "cargotransfer") {
        IntArgBlock save(args);
        callPlanetMethod(pl, univ, ipmCargoTransfer, serverXC, args);
        result.push_back(new ServerPlanetWriter(pid));
        save.getNext(); // skip cargospec
        if (IntIntValue* iv = dynamic_cast<IntIntValue*>(save.getNext()))
            if (getDisplayedTurn().getCurrentUniverse().isValidShipId(iv->getValue()))
                result.push_back(new ServerShipWriter(iv->getValue()));
    } else {
        throw IntError("Invalid command");
    }
}

/************************ ServerShipCommandHandler ***********************/

ServerShipCommandHandler::ServerShipCommandHandler(int sid)
    : sid(sid)
{ }

void
ServerShipCommandHandler::processCommand(string_t cmd, IntArgBlock& args, ptr_vector<ServerObjectWriter>& result)
{
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    GShip& sh = univ.getShip(sid);
    // Missing: mark, unmark, fixship, recycleship
    if (cmd == "setcomment") {
        callShipMethod(sh, univ, ismSetComment, serverXC, args);
        result.push_back(new ServerShipWriter(sid));
    } else if (cmd == "setfcode") {
        args.checkArgumentCount(1);
        setShipProperty(sh, univ, ispFCode, args.getNext());
        result.push_back(new ServerShipWriter(sid));
    } else if (cmd == "setname") {
        args.checkArgumentCount(1);
        setShipProperty(sh, univ, ispName, args.getNext());
        result.push_back(new ServerShipxyWriter());
    } else if (cmd == "setwaypoint") {
        callShipMethod(sh, univ, ismSetWaypoint, serverXC, args);
        result.push_back(new ServerShipWriter(sid));
    } else if (cmd == "setenemy") {
        args.checkArgumentCount(1);
        setShipProperty(sh, univ, ispEnemyId, args.getNext());
        result.push_back(new ServerShipWriter(sid));
    } else if (cmd == "setspeed") {
        args.checkArgumentCount(1);
        setShipProperty(sh, univ, ispSpeedId, args.getNext());
        result.push_back(new ServerShipWriter(sid));
    } else if (cmd == "setmission") {
        callShipMethod(sh, univ, ismSetMission, serverXC, args);
        result.push_back(new ServerShipWriter(sid));
    } else if (cmd == "cargotransfer") {
        IntArgBlock save(args);
        callShipMethod(sh, univ, ismCargoTransfer, serverXC, args);
        result.push_back(new ServerShipWriter(sid));
        save.getNext(); // skip cargospec
        if (IntIntValue* iv = dynamic_cast<IntIntValue*>(save.getNext()))
            if (getDisplayedTurn().getCurrentUniverse().isValidShipId(iv->getValue()))
                result.push_back(new ServerShipWriter(iv->getValue()));
    } else if (cmd == "cargoupload") {
        callShipMethod(sh, univ, ismCargoUpload, serverXC, args);
        result.push_back(new ServerShipWriter(sid));
        if (sh.getOrbitPlanetId())
            result.push_back(new ServerPlanetWriter(sh.getOrbitPlanetId()));
    } else if (cmd == "cargounload") {
        callShipMethod(sh, univ, ismCargoUnload, serverXC, args);
        result.push_back(new ServerShipWriter(sid));
        if (sh.getOrbitPlanetId())
            result.push_back(new ServerPlanetWriter(sh.getOrbitPlanetId()));
    } else {
        throw IntError("Invalid command");
    }
}
